/**
 * This code executes a reduction addition 
 * using a binary tree
 * This is intended to be parallelized with threads and tasks.
 * Mexico 2018
 * Jimmy Aguilar Mena
 */

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

#include <sys/types.h>
#include <sys/time.h>

//! Random Number Generator (macro)
#define frand()((double)rand()/(RAND_MAX))

//! Time measure
double mtime()
{
	struct timeval tmp;
	gettimeofday( &tmp, (struct timezone *)0 );
	return tmp.tv_sec + ((double)tmp.tv_usec)/1000000.0;
}

//! Serial reduction
double serial_reduction(const double *array, size_t size)
{
	double result = 0.0;

	for (size_t i = 0; i < size; ++i) {
		result += array[i];
	}
	return result;
}

//! Recursive reduction
double recursive_reduction(const double *array, size_t size)
{
	if (size == 1)
		return array[0];

	double result = 0.0;
	size_t half = size / 2;

	result = recursive_reduction(array, half);
	result += recursive_reduction(&array[half], size - half);
	return result;
}

int main(int argc, char **argv)
{
	// Check command line arguments
	if (argc <2) {
		printf("Usage: %s dimension\n",argv[0]);
		exit(EXIT_FAILURE);
	}
	const size_t elements = atoi(argv[1]);
	double t, r;

	// Allocate array
	double *array = (double *) malloc(elements * sizeof(double));
	if (!array){
		perror("Memory allocation failed\n");
		return -1;
	}

	// Fill array with random numbers this can be improved ;)
	for (size_t i = 0; i < elements; ++i)
		array[i] = frand();

	// Serial reduction
	t = mtime();
	r = serial_reduction(array, elements);
	t = mtime() - t;
	printf("Serial reduction: %lf t = %lf sec\n", r, t);

	// Recursive reduction
	t = mtime();
	r = recursive_reduction(array, elements);
	t = mtime() - t;
	printf("Recursive reduction: %lf t = %lf sec\n", r, t);

	// Release memory
	free(array);
	return 0;
}
